## Excel 倒立振子　DDQN 訓練
## (ゼロから作るDeep Leraning　(斎藤康毅氏 )内の関数を利用させていただきました)
# coding: utf-8
#pyinstaller aipole.py --onefile
import sys, os
import time
sys.path.append(os.pardir)  # 親ディレクトリのファイルをインポートするための設定
import numpy as np
import pickle

class Relu:
    def __init__(self):
        self.mask = None

    def forward(self, x):
        self.mask = (x <= 0)
        out = x.copy()
        out[self.mask] = 0

        return out

    def backward(self, dout):
        dout[self.mask] = 0
        dx = dout

        return dx

class SGD:

    """確率的勾配降下法（Stochastic Gradient Descent）"""

    def __init__(self, lr=0.01):
        self.lr = lr
        
    def update(self, params, grads):
        for key in params.keys():
            params[key] -= self.lr * grads[key] 

class Adam:

    """Adam (http://arxiv.org/abs/1412.6980v8)"""

    def __init__(self, lr=0.001, beta1=0.9, beta2=0.999):
        self.lr = lr
        self.beta1 = beta1
        self.beta2 = beta2
        self.iter = 0
        self.m = None
        self.v = None
        
    def update(self, params, grads):
        if self.m is None:
            self.m, self.v = {}, {}
            for key, val in params.items():
                self.m[key] = np.zeros_like(val)
                self.v[key] = np.zeros_like(val)
        self.iter += 1
        lr_t  = self.lr * np.sqrt(1.0 - self.beta2**self.iter) / (1.0 - self.beta1**self.iter)         
        
        for key in params.keys():
            self.m[key] += (1 - self.beta1) * (grads[key] - self.m[key])
            self.v[key] += (1 - self.beta2) * (grads[key]**2 - self.v[key])
            
            params[key] -= lr_t * self.m[key] / (np.sqrt(self.v[key]) + 1e-7)
            
class Affine:
    def __init__(self, W, b):
        self.W =W
        self.b = b
        
        self.x = None
        self.original_x_shape = None
        # 重み・バイアスパラメータの微分
        self.dW = None
        self.db = None

    def forward(self, x):
        # テンソル対応
        self.original_x_shape = x.shape
        x = x.reshape(x.shape[0], -1)
        self.x = x

        out = np.dot(self.x, self.W) + self.b

        return out

    def backward(self, dout):
        dx = np.dot(dout, self.W.T)
        self.dW = np.dot(self.x.T, dout)
        self.db = np.sum(dout, axis=0)
        
        dx = dx.reshape(*self.original_x_shape)  # 入力データの形状に戻す（テンソル対応）
        return dx

class Squarederor: #岩崎作成
    def __init__(self):
        self.loss = None
        self.y = None 
        self.t = None # 教師データ

    def forward(self, x, t):
        self.t = t
        self.y =x
        batch_size = self.t.shape[0]
        self.loss = 1/2*sum((self.y-self.t)**2)
        
        return self.loss

    def backward(self):
        batch_size = self.t.shape[0]
#        dx = self.y.copy()
        dout=np.clip(self.y-self.t,-1,1) #上限1
        dx =dout #/ batch_size
       
        return dx

class SimpleNet:
    def __init__(self, input_size=4, 
                 hidden_size=32, output_size=2, weight_init_std=0.02):

        # 重みの初期化
        self.params = {}
        self.params['W1'] = weight_init_std * \
                            np.random.randn(input_size, hidden_size)
        self.params['b1'] = np.zeros(hidden_size)
        self.params['W2'] = weight_init_std * \
                            np.random.randn(hidden_size, hidden_size)
        self.params['b2'] = np.zeros(hidden_size)
        self.params['W3'] = weight_init_std * \
                            np.random.randn(hidden_size, output_size)
        self.params['b3'] = np.zeros(output_size)

        # レイヤの生成
        self.layers ={} #OrderedDict()
        self.layers['Affine1'] = Affine(self.params['W1'], self.params['b1'])
        self.layers['Relu1'] = Relu()
        self.layers['Affine2'] = Affine(self.params['W2'], self.params['b2'])
        self.layers['Relu2'] = Relu()
        self.layers['Affine3'] = Affine(self.params['W3'], self.params['b3'])

        self.last_layer = Squarederor() #二乗誤差

    def predict(self, x):
        for layer in self.layers.values():
            x = layer.forward(x)

        return x

    def loss(self, x, t):
        """損失関数を求める
        引数のxは入力データ、tは教師ラベル
        """
        y = self.predict(x)
        return self.last_layer.forward(y, t)


    def gradient(self, x, t):
        """勾配を求める（誤差逆伝搬法）

        Parameters
        ----------
        x : 入力データ
        t : 教師ラベル

        Returns
        -------
        各層の勾配を持ったディクショナリ変数
            grads['W1']、grads['W2']、...は各層の重み
            grads['b1']、grads['b2']、...は各層のバイアス
        """
        # forward
        self.loss(x, t)

        # backward
#        dout = 1
        dout = self.last_layer.backward()

        layers = list(self.layers.values())
        layers.reverse()
        for layer in layers:
            dout = layer.backward(dout)

        # 設定
        grads = {}
        grads['W1'], grads['b1'] = self.layers['Affine1'].dW, self.layers['Affine1'].db
        grads['W2'], grads['b2'] = self.layers['Affine2'].dW, self.layers['Affine2'].db
        grads['W3'], grads['b3'] = self.layers['Affine3'].dW, self.layers['Affine3'].db

        return grads
        
    def save_params(self, file_name=""):
        params = {}
        for key, val in self.params.items():
            params[key] = val
        with open(file_name, 'wb') as f:
            pickle.dump(params, f)

    def load_params(self, file_name=""):
        with open(file_name, 'rb') as f:
            params = pickle.load(f)
        for key, val in params.items():
            self.params[key] = val

        for i, key in enumerate(['Affine1', 'Affine2', 'Affine3']):
            self.layers[key].W = self.params['W' + str(i+1)]
            self.layers[key].b = self.params['b' + str(i+1)]

def paracopy(): #excelに重みを渡す
    #SimpleNet->SimpleNet2へ w bコピー
    for i, key in enumerate(['Affine1', 'Affine2', 'Affine3']):
        network2.layers[key].W = network.layers[key].W
        network2.layers[key].b = network.layers[key].b

            

#---------------------------Trainer --------------------------------------

class Trainer:
    """ニューラルネットの訓練を行うクラス
    """
    def __init__(self, network,  
                 epochs=20, mini_batch_size=100,
                 optimizer='SGD', optimizer_param={'lr':0.01}, 
                 evaluate_sample_num_per_epoch=None, verbose=True):
        self.network = network
        self.verbose = verbose
        self.epochs = epochs
        self.batch_size = mini_batch_size
        self.evaluate_sample_num_per_epoch = evaluate_sample_num_per_epoch

        # optimizer
        optimizer_class_dict = {'sgd':SGD, 'adam':Adam}
        self.optimizer = optimizer_class_dict[optimizer.lower()](**optimizer_param)
        

    def train_step(self):
        batch_mask = np.random.choice(self.train_size, self.batch_size)
        x_batch1 = self.x_train1[batch_mask] #現在の状態
        x_batch2 = self.x_train2[batch_mask] #次の状態
        acth = self.t_train[batch_mask] #力　報酬　終了
        t_batch=self.network.predict(x_batch1) #現在の価値を読み込んでおく
        idx=np.arange(len(t_batch))
        nval=self.network.predict(x_batch2) #次の価値 MAINネットワークより
        c=np.argmax(nval,axis=1) #大きい方のインデックス
        nvalt=network2.predict(x_batch2) #次の価値 TARGETネットワーク2より
        mdat=nvalt[idx,c] #次の価値 TARGETネットワーク2より MAIN大きい方
        act=((np.sign(acth[:,0])+1)//2).astype(np.int32) #tb1[,0]は行動 左右
        ## 終了⇒次の状態無し　mdat=0とする
        mdat=mdat*(1-acth[:,2]) #終了の場合消す
        t_batch[idx,act]= (acth[idx,1]+gamma*mdat) #更新する価値

        grads = self.network.gradient(x_batch1, t_batch)
        self.optimizer.update(self.network.params, grads)
        ## パラメータ更新
        loss = self.network.loss(x_batch1, t_batch)
        if self.current_iter % 50==0 : print(str(self.current_iter)+ " train loss:" + str(loss))   
        if self.current_iter % self.iter_per_epoch == 0:
            self.current_epoch += 1

        self.current_iter += 1

    def train(self,x_train1,x_train2, t_train):
        self.x_train1 = x_train1
        self.x_train2 = x_train2
        self.t_train = t_train
        self.train_size = x_train1.shape[0]
        self.iter_per_epoch = max(self.train_size // self.batch_size, 1)
        self.max_iter = int(self.epochs * self.iter_per_epoch)
        self.current_iter = 0
        self.current_epoch = 0

        for i in range(self.max_iter):
            self.train_step()
            ## SimpleNet->SimpleNet2へ w bコピー
            if i%2==0 :
                paracopy()

#---- 配列テキスト出力---追加-岩崎---------
def outh(h, ndimz):
    da = ""
    for j in np.arange(ndimz - 2):
        da = da + str(int(h[j])) + ","
    return da

def hup(mh, h, ndimz):
    up = -1  #桁上がりなし
    #For j = ndim - 1 To 0 Step -1
    for j in np.arange(ndimz-1,-1,-1):
        h[j] = h[j] + 1
        if h[j] > mh[j] - 1:
            h[j] = 0
            up = j #最後に桁上がりした次元 0で終了
            #print("up j="+ str(j)+" "+ str(up))  
        else:
            #print("break j="+ str(j)+" "+ str(up))  
            break
    return up    
def ich(mh, h, ndim, xh):
    dat = 0
    for j in np.arange(ndim - 1,-1,-1):
        dat = dat + h[j] * xh[j]
        if j > 0 :
            xh[j - 1] = xh[j] * mh[j]
    return dat

def shapem(a):
    if type(a) is np.ndarray :
        
        ds=""
        for i in range(a.ndim):
            ds=ds + str(a.shape[i])+','
    else:
        ds=""
    return ds

def dfile(path_w,com,dat): #配列のファイル出力
    ms=shapem(dat)
    if ms=="" :
        with open(path_w, mode='a') as f:
            f.write(com+'(値'+str(dat)+')\n') 
    else:  
        ndimz=dat.ndim
        mh=np.zeros(ndimz)
        for i in np.arange(ndimz):
            mh[i]=dat.shape[i]
        h=np.zeros(ndimz)
        xh=np.zeros(ndimz)
        xh[ndimz - 1]=1
        up=-1
        pr=""
        dd=dat.flatten()
        with open(path_w, mode='a') as f:
            f.write(com+'('+ms+')\n') 
            while up !=0:
                jj=int(ich(mh, h, ndimz, xh))
                #print("jj="+ str(jj))
                pr=pr + " " + str(dd[jj])
                up=hup(mh, h, ndimz) #up=ndim-1 1次元up改行  その他  配列表示
                if up >= 0 : #up == ndim - 1:
                    f.write(pr+'\n')
                    pr=""
                    if up<ndimz - 1 and up!=0 :
                        f.write(  ' <<' + outh(h, ndimz)+'\n')    


def pararead():
    # ---データセットの読み込み-変更　岩崎----
    #  現在の状態 | 次の状態 | 　力, 報酬, 終了
    #  例　　1,2,3,4|2,3,4,5|10,-1,0
    with open(inf) as f:
        data= f.readlines()
        x_train1=[]
        x_train2=[]
        t_train=[]
        for i,da in enumerate(data):
            idx = da.find('|')
            idx2 = da.find('|',idx+1)
            x_train1.append(list(da[0:idx].split(",")))
            x_train2.append(list(da[idx+1:idx2].split(",")))
            t_train.append(list(da[idx2+1:-1].split(",")))
    x_train1=np.array(x_train1).astype(np.float64)  #現在の状態
    x_train2=np.array(x_train2).astype(np.float64) #次の状態
    t_train=np.array(t_train).astype(np.float64) #力 報酬 終了
#    print(x_train1.shape)
#    print(t_train.shape)
    return x_train1,x_train2,t_train

    #  ----バラメータテキスト出力 ---追加　岩崎----------------
def paraout():
    try:
            os.remove(ouf) #ファイル削除
    except:
        print("ファイル無し")

    for key, val in network.params.items():
        dfile(ouf,'i',key)
        dfile(ouf,'params',val)


## デバック用データ削除
try:
    os.remove(fpara) #ファイル削除
except:
    print("デバックファイル無し")

#------処理開始  --------      
max_epochs=20 # 20回　実行
gamma=0.95 #次の価値に掛ける係数

ppath = os.getcwd() #プログラムのホルダ
inf=ppath+"\pole.txt" # excelからの訓練データ
ouf=ppath+"\polepara.txt" #python重みexcelへの出力ファイル名
kaishi=ppath+"\polest.txt" #Excelからの開始指示
shuryo=ppath+"\poleen.txt" #Excelへの終了報告
fpara=ppath+"\poledat.txt" #デバックデータ出力ファイル名 
pypara=ppath+"\params.pkl" #Python重み保存 デバック用
# パラメータの読み込み
print("処理開始")

# 処理依頼が来るまで待つ
while True: ## 繰り返し
    ok=0
    print("開始指示待ち")
    while ok==0:
        try:
            with open(kaishi) as f:
                l = f.readlines()
            ok=1
            os.remove(kaishi) #依頼ファイル削除
        except:
            ok=0
            time.sleep(0.1)
#            break

    if l[0].find("初期化")>=0:
        ##ネットワーク初期化
        network = SimpleNet(input_size=4, 
                                hidden_size=32, output_size=2, weight_init_std=0.02) #メインネットワーク
        network2 = SimpleNet(input_size=4, 
                                hidden_size=32, output_size=2, weight_init_std=0.02) #ターゲットネットワーク
        x_train1=[]
        x_train2=[]
        t_train=[]

        trainer = Trainer(network, 
                          epochs=max_epochs, mini_batch_size=100,
                          optimizer='Adam', optimizer_param={'lr': 0.00005},
                          evaluate_sample_num_per_epoch=1000)

        print(">>>>初期化完了")
###  デバック　確認用
    if l[0].find("確認")>=0:
        network.load_params(pypara)
        print("パラメータ読み込み完了")
        x_train1, x_train2, t_train=pararead() #excelからのパラメータファイル読み込み
        tb2=network.predict(x_train1)
        dfile(fpara,'結果tb2',tb2)
        print("データ出力完了")
        break
    print("データ読み込み")
    x_train1, x_train2, t_train=pararead() #excelからのパラメータファイル読み込み
    trainer.train(x_train1,x_train2, t_train)

    paraout() #python重みファイル　excelへ出力
    
    #終了書き込み
    with open(shuryo, mode='a') as f:
        f.write('end') 

# パラメータの保存
#    network.save_params(pypara) #デバック用
#    print("訓練終了")
#    break
